"use strict";
// Copyright 2020 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
Object.defineProperty(exports, "__esModule", { value: true });
exports.TerraformModule = void 0;
const release_pr_1 = require("../release-pr");
const conventional_commits_1 = require("../conventional-commits");
const checkpoint_1 = require("../util/checkpoint");
// Generic
const changelog_1 = require("../updaters/changelog");
// Terraform specific.
const readme_1 = require("../updaters/terraform/readme");
class TerraformModule extends release_pr_1.ReleasePR {
    async _run() {
        const latestTag = await this.gh.latestTag();
        const commits = await this.commits(latestTag ? latestTag.sha : undefined);
        const cc = new conventional_commits_1.ConventionalCommits({
            commits,
            githubRepoUrl: this.repoUrl,
            bumpMinorPreMajor: this.bumpMinorPreMajor,
        });
        const candidate = await this.coerceReleaseCandidate(cc, latestTag);
        const changelogEntry = await cc.generateChangelogEntry({
            version: candidate.version,
            currentTag: `v${candidate.version}`,
            previousTag: candidate.previousTag,
        });
        // don't create a release candidate until user facing changes
        // (fix, feat, BREAKING CHANGE) have been made; a CHANGELOG that's
        // one line is a good indicator that there were no interesting commits.
        if (this.changelogEmpty(changelogEntry)) {
            checkpoint_1.checkpoint(`no user facing commits found since ${latestTag ? latestTag.sha : 'beginning of time'}`, checkpoint_1.CheckpointType.Failure);
            return;
        }
        const updates = [];
        updates.push(new changelog_1.Changelog({
            path: 'CHANGELOG.md',
            changelogEntry,
            version: candidate.version,
            packageName: this.packageName,
        }));
        updates.push(new readme_1.ReadMe({
            path: 'README.md',
            changelogEntry,
            version: candidate.version,
            packageName: this.packageName,
        }));
        await this.openPR(commits[0].sha, `${changelogEntry}\n---\n`, updates, candidate.version);
    }
    defaultInitialVersion() {
        return '0.1.0';
    }
}
exports.TerraformModule = TerraformModule;
TerraformModule.releaserName = 'terraform-module';
//# sourceMappingURL=terraform-module.js.map